;;;; MaiSQL --- Common Lisp Interface Layer to SQL Databases
;;;; This is copyrighted software.  See documentation for terms.
;;;; 
;;;; postgresql.cl --- FFI interface to PostgreSQL on Unix
;;;; 
;;;; Checkout Tag: $Name: debian_version_1_2_1-1 $
;;;; $Id: glue-cmucl.lisp,v 1.3 2002/12/05 17:37:14 jesse Exp $

(in-package :MAISQL-POSTGRESQL)

(eval-when (:compile-toplevel :load-toplevel)

  #+cmu
  (defmacro with-lock (lock &body body)
    `(mp:with-lock-held ,lock
      ,@body))
  #-cmu
  (defmacro with-lock (lock &body body)
    `(progn
      ,@body))  

(defclass postgresql-database (database)
  ((conn-ptr :accessor database-conn-ptr :initarg :conn-ptr
	     :type psql-conn-ptr)
   (lock
    :accessor database-lock
    #+(or cmu lispworks)
    :initform
    #+(or cmu lispworks) (mp:make-lock #-cmu :name "connection lock"))))

)

(defun %pg-database-connection (connection-spec)
  (check-connection-spec connection-spec :postgresql
			 (host db user password &optional port options tty))
  (destructuring-bind (host db user password &optional port options tty)
      connection-spec
    (let ((connection (pqsetdblogin host port options tty db user password)))
      (declare (type psql-conn-ptr connection))
      (unless (eq (pqstatus connection) :connection-ok)
	;; Connect failed
	(error 'maisql-connect-error
	       :database-type :postgresql
	       :connection-spec connection-spec
	       :errno (pqstatus connection)
	       :error (pqerrormessage connection)))
      connection)))  

(defmethod database-reconnect ((database postgresql-database))
    (let ((foo (database-lock database)))
      (with-lock (foo)
	(with-slots (maisql-sys::spec conn-ptr)
	    database
	  (setf conn-ptr (%pg-database-connection maisql-sys::spec))
	  database))))

(defun %trim-crlf (string)
  (string-right-trim '(#\Return #\Newline) string))

(defun %pg-query (database query-expression)
  (let ((retry nil)
        (result nil)
        (conn-ptr (database-conn-ptr database)))
    (declare (type (alien psql-conn) conn-ptr))
    (tagbody
     execute
       (setq result (pqexec conn-ptr query-expression))
       (when (null-alien result)
         (when (null retry)
           (setq retry t)
           (database-reconnect database)
           (setq conn-ptr (database-conn-ptr database))
           (go execute))
         (error 'maisql-sql-error
                :database database
                :expression query-expression
                :errno nil
                :error (%trim-crlf (pqerrormessage conn-ptr)))))
    result))

(defun %pg-exec (database sql-expression)
  (let ((conn-ptr (database-conn-ptr database))
        (retry nil)
        (result nil))
    (declare (type (alien psql-conn) conn-ptr))
    (tagbody
     execute
       (setq result (pqexec conn-ptr sql-expression))
       (when (null-alien result)
         (when (null retry)
           (setq retry t)
           (database-reconnect database)
           (setq conn-ptr (database-conn-ptr database))
           (go execute))
         (error 'maisql-sql-error
                :database database
                :expression sql-expression
                :errno nil
                :error (%trim-crlf (pqerrormessage conn-ptr)))))
    result))

(defun %pg-query-result-set (database query-expression)
  (let ((conn-ptr (database-conn-ptr database)))
    (declare (type (alien psql-conn) conn-ptr))
    (let ((result (pqexec conn-ptr query-expression)))
      (when (null-alien result)
	(error 'maisql-sql-error
	       :database database
	       :expression query-expression
	       :errno nil
	       :error "PQexec failed to return result structure!"))
      result)))
